@@ -2,17 +2,13 @@
 switchboard.tests.test_manager
 ~~~~~~~~~~~~~~~
 
-:copyright: (c) 2012 SourceForge.
+:copyright: (c) 2015 Kyle Adams.
 :license: Apache License 2.0, see LICENSE for more details.
 """
+import threading
 
-from nose.tools import (
-    assert_equals,
-    assert_true,
-    assert_false,
-    assert_raises
-)
-from mock import patch
+import pytest
+from unittest.mock import Mock, patch
 from webob import Request
 from webob.exc import HTTPNotFound, HTTPFound
 
@@ -26,13 +22,14 @@ from ..decorators import switch_is_active
 from ..models import (
     Switch,
     SELECTIVE, DISABLED, GLOBAL, INHERIT,
+    INCLUDE, EXCLUDE
 )
-from ..manager import SwitchManager
+from ..manager import registry, SwitchManager
 from ..helpers import MockCollection
 from ..settings import settings
 
 
-class TestAPI(object):
+class TestAPI:
     def setup(self):
         settings.SWITCHBOARD_SWITCH_DEFAULTS = {
             'active_by_default': {
@@ -55,47 +52,28 @@ class TestAPI(object):
         Switch.c.drop()
 
     def test_builtin_registration(self):
-        assert_true('switchboard.builtins.QueryStringConditionSet'
-                    in self.operator._registry)
-        assert_true('switchboard.builtins.IPAddressConditionSet'
-                    in self.operator._registry)
-        assert_true('switchboard.builtins.HostConditionSet'
-                    in self.operator._registry)
-        assert_equals(len(list(self.operator.get_condition_sets())), 3,
-                      self.operator)
+        assert 'switchboard.builtins.QueryStringConditionSet' in registry
+        assert 'switchboard.builtins.IPAddressConditionSet' in registry
+        assert 'switchboard.builtins.HostConditionSet' in registry
+        assert len(list(self.operator.get_condition_sets())) == 3, self.operator
 
-    @patch('switchboard.signals.switch_checked.send')
-    def test_switch_checked_signal(self, send):
-        condition_set = 'switchboard.builtins.IPAddressConditionSet'
-        key = 'test'
+    def test_unregister(self):
+        self.operator.unregister(QueryStringConditionSet)
+        condition_set_id = 'switchboard.builtins.QueryStringConditionSet'
+        assert not condition_set_id in registry
+        assert len(list(self.operator.get_condition_sets())) == 2, self.operator
 
-        switch = Switch.create(
-            key=key,
-            status=SELECTIVE,
-        )
-        switch = self.operator[key]
-        req = Request.blank('/')
-
-        result = self.operator.is_active(key, req)
-        assert_false(result)
-        send.assert_called_with(key, active=result)
-
-        switch.add_condition(
-            condition_set=condition_set,
-            field_name='ip_address',
-            condition='192.168.1.1',
-        )
-        req.environ['REMOTE_ADDR'] = '192.168.1.1'
-
-        result = self.operator.is_active(key, req)
-        assert_true(result)
-        send.assert_called_with(key, active=result)
+    def test_get_all_conditions(self):
+        conditions = list(self.operator.get_all_conditions())
+        assert len(conditions) == 5
+        for set_id, label, field in conditions:
+            assert set_id in registry
 
     @patch('switchboard.base.MongoModelDict.get_default')
     def test_error(self, get_default):
         # force the is_active call to fail right away
         get_default.side_effect = Exception('Boom!')
-        assert_false(self.operator.is_active('test'))
+        assert not self.operator.is_active('test')
 
     def test_exclusions(self):
         condition_set = 'switchboard.builtins.IPAddressConditionSet'
@@ -120,10 +98,10 @@ class TestAPI(object):
 
         req = Request.blank('/')
         req.environ['REMOTE_ADDR'] = '192.168.1.1'
-        assert_true(self.operator.is_active('test', req))
+        assert self.operator.is_active('test', req)
 
         req.environ['REMOTE_ADDR'] = '10.1.1.1'
-        assert_false(self.operator.is_active('test', req))
+        assert not self.operator.is_active('test', req)
 
     def test_only_exclusions(self):
         condition_set = 'switchboard.builtins.IPAddressConditionSet'
@@ -143,10 +121,10 @@ class TestAPI(object):
 
         req = Request.blank('/')
         req.environ['REMOTE_ADDR'] = '192.168.1.1'
-        assert_false(self.operator.is_active('test', req))
+        assert not self.operator.is_active('test', req)
 
         req.environ['REMOTE_ADDR'] = '10.1.1.1'
-        assert_false(self.operator.is_active('test', req))
+        assert not self.operator.is_active('test', req)
 
     def test_decorator_for_ip_address(self):
         condition_set = 'switchboard.builtins.IPAddressConditionSet'
@@ -159,14 +137,13 @@ class TestAPI(object):
 
         req = Request.blank('/')
         req.environ['REMOTE_ADDR'] = '192.168.1.1'
-        original_func = self.operator.get_request
-        self.operator.get_request = lambda: req
+        self.operator.context['request'] = req
 
         @switch_is_active('test', operator=self.operator)
         def test():
             return True
 
-        assert_raises(HTTPNotFound, test)
+        pytest.raises(HTTPNotFound, test)
 
         switch.status = SELECTIVE
         switch.save()
@@ -177,7 +154,7 @@ class TestAPI(object):
             condition='192.168.1.1',
         )
 
-        assert_true(test())
+        assert test()
 
         # add in a second condition, so that removing the first one won't kick
         # in the "no conditions returns is_active True for selective switches"
@@ -193,7 +170,7 @@ class TestAPI(object):
             condition='192.168.1.1',
         )
 
-        assert_raises(HTTPNotFound, test)
+        pytest.raises(HTTPNotFound, test)
 
         switch.add_condition(
             condition_set=condition_set,
@@ -201,7 +178,7 @@ class TestAPI(object):
             condition='192.168.1.1',
         )
 
-        assert_true(test())
+        assert test()
 
         switch.clear_conditions(
             condition_set=condition_set,
@@ -214,7 +191,7 @@ class TestAPI(object):
             condition='50-100',
         )
 
-        assert_true(test())
+        assert test()
 
         switch.clear_conditions(
             condition_set=condition_set,
@@ -226,8 +203,7 @@ class TestAPI(object):
             condition='0-50',
         )
 
-        assert_raises(HTTPNotFound, test)
-        self.operator.get_request = original_func
+        pytest.raises(HTTPNotFound, test)
 
     def test_decorator_with_redirect(self):
         Switch.create(
@@ -236,15 +212,13 @@ class TestAPI(object):
         )
 
         req = Request.blank('/')
-        original_func = self.operator.get_request
-        self.operator.get_request = lambda: req
+        self.operator.context['request'] = lambda: req
 
         @switch_is_active('test', redirect_to='/foo')
         def test():
             return True
 
-        assert_raises(HTTPFound, test)
-        self.operator.get_request = original_func
+        pytest.raises(HTTPFound, test)
 
     def test_global(self):
         switch = Switch.create(
@@ -255,14 +229,14 @@ class TestAPI(object):
 
         req = Request.blank('/')
 
-        assert_false(self.operator.is_active('test'))
-        assert_false(self.operator.is_active('test', req))
+        assert not self.operator.is_active('test')
+        assert not self.operator.is_active('test', req)
 
         switch.status = GLOBAL
         switch.save()
 
-        assert_true(self.operator.is_active('test'))
-        assert_true(self.operator.is_active('test', req))
+        assert self.operator.is_active('test')
+        assert self.operator.is_active('test', req)
 
     def test_disable(self):
         switch = Switch.create(key='test')
@@ -274,20 +248,20 @@ class TestAPI(object):
         switch.status = DISABLED
         switch.save()
 
-        assert_false(self.operator.is_active('test'))
+        assert not self.operator.is_active('test')
 
-        assert_false(self.operator.is_active('test', req))
+        assert not self.operator.is_active('test', req)
 
     def test_deletion(self):
         switch = Switch.create(key='test')
 
         switch = self.operator['test']
 
-        assert_true('test' in self.operator)
+        assert 'test' in self.operator
 
         switch.delete()
 
-        assert_false('test' in self.operator)
+        assert not 'test' in self.operator
 
     def test_expiration(self):
         switch = Switch.create(key='test')
@@ -297,7 +271,7 @@ class TestAPI(object):
         switch.status = DISABLED
         switch.save()
 
-        assert_false(self.operator.is_active('test'))
+        assert not self.operator.is_active('test')
 
     def test_ip_address_internal_ips(self):
         condition_set = 'switchboard.builtins.IPAddressConditionSet'
@@ -311,7 +285,7 @@ class TestAPI(object):
         req = Request.blank('/')
         req.environ['REMOTE_ADDR'] = '192.168.1.1'
 
-        assert_false(self.operator.is_active('test', req))
+        assert not self.operator.is_active('test', req)
 
         switch.add_condition(
             condition_set=condition_set,
@@ -321,11 +295,11 @@ class TestAPI(object):
 
         settings.SWITCHBOARD_INTERNAL_IPS = ['192.168.1.1']
 
-        assert_true(self.operator.is_active('test', req))
+        assert self.operator.is_active('test', req)
 
         settings.SWITCHBOARD_INTERNAL_IPS = []
 
-        assert_false(self.operator.is_active('test', req))
+        assert not self.operator.is_active('test', req)
 
     def test_ip_address(self):
         condition_set = 'switchboard.builtins.IPAddressConditionSet'
@@ -339,7 +313,7 @@ class TestAPI(object):
         req = Request.blank('/')
         req.environ['REMOTE_ADDR'] = '192.168.1.1'
 
-        assert_false(self.operator.is_active('test', req))
+        assert not self.operator.is_active('test', req)
 
         switch.add_condition(
             condition_set=condition_set,
@@ -347,7 +321,7 @@ class TestAPI(object):
             condition='192.168.1.1',
         )
 
-        assert_true(self.operator.is_active('test', req))
+        assert self.operator.is_active('test', req)
 
         switch.clear_conditions(
             condition_set=condition_set,
@@ -358,13 +332,13 @@ class TestAPI(object):
             condition='127.0.0.1',
         )
 
-        assert_false(self.operator.is_active('test', req))
+        assert not self.operator.is_active('test', req)
 
         switch.clear_conditions(
             condition_set=condition_set,
         )
 
-        assert_false(self.operator.is_active('test', req))
+        assert not self.operator.is_active('test', req)
 
         switch.add_condition(
             condition_set=condition_set,
@@ -372,21 +346,7 @@ class TestAPI(object):
             condition='50-100',
         )
 
-        assert_true(self.operator.is_active('test', req))
-
-        # test with mock request
-        req = self.operator.as_request(ip_address='192.168.1.1')
-        assert_true(self.operator.is_active('test', req))
-
-        switch.clear_conditions(
-            condition_set=condition_set,
-        )
-        switch.add_condition(
-            condition_set=condition_set,
-            field_name='percent',
-            condition='0-50',
-        )
-        assert_false(self.operator.is_active('test', req))
+        assert self.operator.is_active('test', req)
 
     def test_to_dict(self):
         condition_set = 'switchboard.builtins.IPAddressConditionSet'
@@ -407,35 +367,35 @@ class TestAPI(object):
 
         result = switch.to_dict(self.operator)
 
-        assert_true('label' in result)
-        assert_equals(result['label'], 'my switch')
+        assert 'label' in result
+        assert result['label'] == 'my switch'
 
-        assert_true('status' in result)
-        assert_equals(result['status'], SELECTIVE)
+        assert 'status' in result
+        assert result['status'] == SELECTIVE
 
-        assert_true('description' in result)
-        assert_equals(result['description'], 'foo bar baz')
+        assert 'description' in result
+        assert result['description'] == 'foo bar baz'
 
-        assert_true('key' in result)
-        assert_equals(result['key'], 'test')
+        assert 'key' in result
+        assert result['key'] == 'test'
 
-        assert_true('conditions' in result)
-        assert_equals(len(result['conditions']), 1)
+        assert 'conditions' in result
+        assert len(result['conditions']) == 1
 
         condition = result['conditions'][0]
-        assert_true('id' in condition)
-        assert_equals(condition['id'], condition_set)
-        assert_true('label' in condition)
-        assert_equals(condition['label'], 'IP Address')
-        assert_true('conditions' in condition)
-        assert_equals(len(condition['conditions']), 1)
+        assert 'id' in condition
+        assert condition['id'] == condition_set
+        assert 'label' in condition
+        assert condition['label'] == 'IP Address'
+        assert 'conditions' in condition
+        assert len(condition['conditions']) == 1
 
         inner_condition = condition['conditions'][0]
-        assert_equals(len(inner_condition), 4)
-        assert_true(inner_condition[0], 'ip_address')
-        assert_true(inner_condition[1], '192.168.1.1')
-        assert_true(inner_condition[2], '192.168.1.1')
-        assert_false(inner_condition[3])
+        assert len(inner_condition) == 4
+        assert inner_condition[0], 'ip_address'
+        assert inner_condition[1], '192.168.1.1'
+        assert inner_condition[2], '192.168.1.1'
+        assert not inner_condition[3]
 
     def test_remove_condition(self):
         condition_set = 'switchboard.builtins.IPAddressConditionSet'
@@ -450,7 +410,7 @@ class TestAPI(object):
         req1.environ['REMOTE_ADDR'] = '192.168.1.1'
 
         # inactive if selective with no conditions
-        assert_false(self.operator.is_active('test', req1))
+        assert not self.operator.is_active('test', req1)
 
         req2 = Request.blank('/2')
         req2.environ['REMOTE_ADDR'] = '10.1.1.1'
@@ -460,9 +420,9 @@ class TestAPI(object):
             field_name='internal_ip',
             condition='1',
         )
-        assert_true(self.operator.is_active('test', req2))
+        assert self.operator.is_active('test', req2)
         # No longer is_active for IP as we have other conditions
-        assert_false(self.operator.is_active('test', req1))
+        assert not self.operator.is_active('test', req1)
 
         switch.remove_condition(
             condition_set=condition_set,
@@ -471,8 +431,8 @@ class TestAPI(object):
         )
 
         # back to inactive for everyone with no conditions
-        assert_false(self.operator.is_active('test', req1))
-        assert_false(self.operator.is_active('test', req2))
+        assert not self.operator.is_active('test', req1)
+        assert not self.operator.is_active('test', req2)
 
     def test_switch_defaults(self):
         """Test that defaults are pulled from SWITCHBOARD_SWITCH_DEFAULTS.
@@ -480,20 +440,18 @@ class TestAPI(object):
         Requires SwitchManager to use auto_create.
 
         """
-        assert_true(self.operator.is_active('active_by_default'))
-        assert_false(self.operator.is_active('inactive_by_default'))
-        assert_equals(
-            self.operator['inactive_by_default'].label,
-            'Default Inactive',
-        )
-        assert_equals(
-            self.operator['active_by_default'].label,
-            'Default Active',
-        )
+        assert self.operator.is_active('active_by_default')
+        assert not self.operator.is_active('inactive_by_default')
+        assert (
+            self.operator['inactive_by_default'].label ==
+            'Default Inactive')
+        assert (
+            self.operator['active_by_default'].label ==
+            'Default Active')
         active_by_default = self.operator['active_by_default']
         active_by_default.status = DISABLED
         active_by_default.save()
-        assert_false(self.operator.is_active('active_by_default'))
+        assert not self.operator.is_active('active_by_default')
 
     def test_invalid_condition(self):
         condition_set = 'switchboard.builtins.IPAddressConditionSet'
@@ -507,7 +465,7 @@ class TestAPI(object):
         req1 = Request.blank('/1')
 
         # inactive if selective with no conditions
-        assert_false(self.operator.is_active('test', req1))
+        assert not self.operator.is_active('test', req1)
 
         req2 = Request.blank('/2')
         req2.environ['REMOTE_ADDR'] = '10.1.1.1'
@@ -517,7 +475,7 @@ class TestAPI(object):
             field_name='foo',
             condition='1',
         )
-        assert_false(self.operator.is_active('test', req2))
+        assert not self.operator.is_active('test', req2)
 
     def test_inheritance(self):
         condition_set = 'switchboard.builtins.IPAddressConditionSet'
@@ -544,30 +502,30 @@ class TestAPI(object):
 
         req = Request.blank('/')
         req.environ['REMOTE_ADDR'] = '1.1.1.1'
-        assert_true(self.operator.is_active('test:child', req))
+        assert self.operator.is_active('test:child', req)
 
         req.environ['REMOTE_ADDR'] = '20.20.20.20'
-        assert_false(self.operator.is_active('test:child', req))
+        assert not self.operator.is_active('test:child', req)
 
         # Test parent with disabled status.
         parent_switch.status = DISABLED
         parent_switch.save()
 
         req.environ['REMOTE_ADDR'] = '1.1.1.1'
-        assert_false(self.operator.is_active('test:child', req))
+        assert not self.operator.is_active('test:child', req)
 
         req.environ['REMOTE_ADDR'] = '20.20.20.20'
-        assert_false(self.operator.is_active('test:child', req))
+        assert not self.operator.is_active('test:child', req)
 
         # Test parent with global status.
         parent_switch.status = GLOBAL
         parent_switch.save()
 
         req.environ['REMOTE_ADDR'] = '1.1.1.1'
-        assert_true(self.operator.is_active('test:child', req))
+        assert self.operator.is_active('test:child', req)
 
         req.environ['REMOTE_ADDR'] = '20.20.20.20'
-        assert_true(self.operator.is_active('test:child', req))
+        assert self.operator.is_active('test:child', req)
 
     def test_parent_override_child_state(self):
         Switch.create(
@@ -580,7 +538,7 @@ class TestAPI(object):
             status=GLOBAL,
         )
 
-        assert_false(self.operator.is_active('test:child'))
+        assert not self.operator.is_active('test:child')
 
     def test_child_state_is_used(self):
         Switch.create(
@@ -593,7 +551,7 @@ class TestAPI(object):
             status=DISABLED,
         )
 
-        assert_false(self.operator.is_active('test:child'))
+        assert not self.operator.is_active('test:child')
 
     def test_parent_override_child_condition(self):
         condition_set = 'switchboard.builtins.IPAddressConditionSet'
@@ -618,12 +576,12 @@ class TestAPI(object):
 
         req = Request.blank('/')
         req.environ['REMOTE_ADDR'] = '192.168.1.1'
-        assert_true(self.operator.is_active('test:child', req))
+        assert self.operator.is_active('test:child', req)
 
         req.environ['REMOTE_ADDR'] = '10.1.1.1'
-        assert_false(self.operator.is_active('test:child', req))
+        assert not self.operator.is_active('test:child', req)
 
-        assert_false(self.operator.is_active('test:child'))
+        assert not self.operator.is_active('test:child')
 
     def test_child_condition_differing_than_parent_loses(self):
         condition_set = 'switchboard.builtins.IPAddressConditionSet'
@@ -656,15 +614,15 @@ class TestAPI(object):
 
         req = Request.blank('/')
         req.environ['REMOTE_ADDR'] = '192.168.1.1'
-        assert_false(self.operator.is_active('test:child', req))
+        assert not self.operator.is_active('test:child', req)
 
         req.environ['REMOTE_ADDR'] = '10.1.1.1'
-        assert_false(self.operator.is_active('test:child', req))
+        assert not self.operator.is_active('test:child', req)
 
         req.environ['REMOTE_ADDR'] = '127.0.0.1'
-        assert_false(self.operator.is_active('test:child', req))
+        assert not self.operator.is_active('test:child', req)
 
-        assert_false(self.operator.is_active('test:child'))
+        assert not self.operator.is_active('test:child')
 
     def test_child_condition_including_parent_wins(self):
         condition_set = 'switchboard.builtins.IPAddressConditionSet'
@@ -702,18 +660,83 @@ class TestAPI(object):
 
         req = Request.blank('/')
         req.environ['REMOTE_ADDR'] = '192.168.1.1'
-        assert_true(self.operator.is_active('test:child', req))
+        assert self.operator.is_active('test:child', req)
 
         req.environ['REMOTE_ADDR'] = '10.1.1.1'
-        assert_false(self.operator.is_active('test:child', req))
+        assert not self.operator.is_active('test:child', req)
 
         req.environ['REMOTE_ADDR'] = '127.0.0.1'
-        assert_false(self.operator.is_active('test:child', req))
+        assert not self.operator.is_active('test:child', req)
 
-        assert_false(self.operator.is_active('test:child'))
+        assert not self.operator.is_active('test:child')
+
+    def test_version_switch_no_user(self):
+        switch = Mock()
+        switch.save_version = Mock()
+        self.operator.context = dict()
+        self.operator.version_switch(switch)
+        switch.save_version.assert_called_with(username='')
+
+    def test_version_switch_user_dict(self):
+        username = 'test'
+        switch = Mock()
+        switch.save_version = Mock()
+        self.operator.context = dict(user=dict(username=username))
+        self.operator.version_switch(switch)
+        switch.save_version.assert_called_with(username=username)
+
+    def test_version_switch_user_class(self):
+        username = 'test'
+        switch = Mock()
+        switch.save_version = Mock()
+        self.operator.context = dict(user=Mock(username=username))
+        self.operator.version_switch(switch)
+        switch.save_version.assert_called_with(username=username)
+
+    def test_version_switch_user_class_properror(self):
+        class OurUser:
+            @property
+            def username(self):
+                raise OSError('whoops')
+        switch = Mock()
+        switch.save_version = Mock()
+        self.operator.context = dict(user=OurUser())
+        self.operator.version_switch(switch)
+        switch.save_version.assert_called_with(username='')
+
+    def test_version_switch_nonuser_class(self):
+        class NonUser:
+            pass
+        switch = Mock()
+        switch.save_version = Mock()
+        self.operator.context = dict(user=NonUser())
+        self.operator.version_switch(switch)
+        switch.save_version.assert_called_with(username='')
+
+    def test_version_switch_nonuser_dict(self):
+        switch = Mock()
+        switch.save_version = Mock()
+        self.operator.context = dict(user=dict())
+        self.operator.version_switch(switch)
+        switch.save_version.assert_called_with(username='')
+
+    def test_version_switch_save_error(self):
+        switch = Mock()
+        switch.save_version = Mock()
+        switch.save_version.side_effect = Exception('Boom!')
+        self.operator.context = dict()
+        # Don't need to assert, just need to make sure things don't explode.
+        self.operator.version_switch(switch)
+
+    @patch('switchboard.base.MongoModelDict.__getitem__')
+    def test_defaults_on_key_error(self, getitem):
+        getitem.side_effect = KeyError()
+        operator = SwitchManager()
+        assert operator.is_active('test', default=True)
+        assert not operator.is_active('test', default=False)
 
 
-class TestConfigure(object):
+class TestConfigure:
     def setup(self):
         self.config = dict(
             mongo_host='mongodb',
@@ -731,27 +754,177 @@ class TestConfigure(object):
         Switch.c = MockCollection()
 
     def assert_settings(self):
-        for k, v in self.config.iteritems():
-            assert_equals(getattr(settings, 'SWITCHBOARD_%s' % k.upper()), v)
+        for k, v in self.config.items():
+            assert getattr(settings, 'SWITCHBOARD_%s' % k.upper()) == v
 
-    @patch('switchboard.manager.Connection')
-    def test_unnested(self, Connection):
-        configure(self.config)
+    @patch('switchboard.manager.MongoClient')
+    def test_unnested(self, MongoClient):
+        configure(self.config, allow_no_mongo=True)
         self.assert_settings()
-        assert_false(isinstance(Switch.c, MockCollection))
+        assert not isinstance(Switch.c, MockCollection)
 
-    @patch('switchboard.manager.Connection')
-    def test_nested(self, Connection):
+    @patch('switchboard.manager.MongoClient')
+    def test_nested(self, MongoClient):
         cfg = {}
-        for k, v in self.config.iteritems():
+        for k, v in self.config.items():
             cfg['switchboard.%s' % k] = v
         cfg['foo.bar'] = 'baz'
-        configure(cfg, nested=True)
+        configure(cfg, nested=True, allow_no_mongo=True)
         self.assert_settings()
-        assert_false(isinstance(Switch.c, MockCollection))
+        assert not isinstance(Switch.c, MockCollection)
 
-    @patch('switchboard.manager.Connection')
-    def test_database_failure(self, Connection):
-        Connection.side_effect = Exception('Boom!')
-        configure(self.config)
-        assert_true(isinstance(Switch.c, MockCollection))
+    @patch('switchboard.manager.MongoClient')
+    def test_database_failure_permitted(self, MongoClient):
+        MongoClient.side_effect = Exception('Boom!')
+        configure(self.config, allow_no_mongo=True)
+        assert isinstance(Switch.c, MockCollection)
+
+    @patch('switchboard.manager.MongoClient')
+    def test_database_failure_fails(self, MongoClient):
+        class CustomException(Exception):
+            pass
+        MongoClient.side_effect = CustomException('Boom!')
+        with pytest.raises(CustomException):
+            configure(self.config)
+
+
+class TestManagerConcurrency:
+
+    def setup(self):
+        self.operator = SwitchManager(auto_create=True)
+        self.exc = None
+
+    def test_context_thread_safety(self):
+        '''
+        Verify that switch contexts are not shared across threads (i.e.,
+        between requests).
+        '''
+        def set_context():
+            self.operator.context['foo'] = 'bar'
+
+        def verify_context():
+            try:
+                assert self.operator.context.get('foo') == None
+            except Exception as e:
+                self.exc = e
+
+        t1 = threading.Thread(target=set_context)
+        t1.start()
+        t1.join()
+        t2 = threading.Thread(target=verify_context)
+        t2.start()
+        t2.join()
+
+        if self.exc:
+            raise self.exc
+
+
+class TestManagerResultCaching:
+
+    def setup(self):
+        self.operator = SwitchManager(auto_create=True)
+        self.operator.result_cache = {}
+
+    def test_all(self):
+        switch = Switch.create(
+            key='test',
+            status=GLOBAL,
+        )
+        # first time
+        assert self.operator.is_active('test')
+        # still the same 2nd time
+        assert self.operator.is_active('test')
+        # still the same even if something actually changed
+        switch.status = DISABLED
+        switch.save()
+        assert self.operator.is_active('test')
+        # changes after cache is cleared
+        self.operator.result_cache = {}
+        assert not self.operator.is_active('test')
+        # make sure the false value was cached too
+        switch.status = GLOBAL
+        switch.save()
+        assert not self.operator.is_active('test')
+
+
+class TestManagerResultCacheDecorator:
+
+    def setup(self):
+        # Gets a pure function, otherwise we get an unbound function that we
+        # can't call.
+        self.with_result_cache = SwitchManager.__dict__['with_result_cache']
+
+        # a simple "is_active" to wrap
+        def is_active_func(self, key, *instances, **kwargs):
+            return True
+
+        # wrap it with the decorator
+        self.cached_is_active_func = self.with_result_cache(is_active_func)
+
+    def test_decorator_nocache(self):
+        operator_self = Mock(result_cache=None)
+        result = self.cached_is_active_func(operator_self, 'mykey')
+        assert result
+        assert operator_self.result_cache == None
+
+    def test_decorator_simple(self):
+        operator_self = Mock(result_cache={})
+        result = self.cached_is_active_func(operator_self, 'mykey')
+        assert result
+        assert operator_self.result_cache == {
+            (('mykey',), ()): True
+        }
+
+    def test_decorator_uses_cache(self):
+        # Put False in cache, to ensure only the cache is used, not
+        # is_active_func.
+        operator_self = Mock(result_cache={
+            (('mykey',), ()): False
+        })
+        result = self.cached_is_active_func(operator_self, 'mykey')
+        assert not result
+        assert operator_self.result_cache == {
+            (('mykey',), ()): False
+        }
+
+    def test_decorator_with_params(self):
+        operator_self = Mock(result_cache={})
+        result = self.cached_is_active_func(operator_self, 'mykey',
+                                            'someval', a=1, b=2)
+        assert result
+        assert operator_self.result_cache == {
+            (('mykey', 'someval'),
+             (('a', 1), ('b', 2))): True
+        }
+
+    def test_decorator_uncachable_params(self):
+        operator_self = Mock(result_cache={})
+        # A dict isn't hashable, can't be cached.
+        result = self.cached_is_active_func(operator_self, 'mykey', {})
+        assert result
+        assert operator_self.result_cache == {}
+
+
+class TestManagerConstants:
+    def setup(self):
+        self.operator = SwitchManager()
+
+    def test_disabled(self):
+        assert hasattr(self.operator, 'DISABLED')
+        assert self.operator.DISABLED == DISABLED
+
+    def test_selective(self):
+        assert hasattr(self.operator, 'SELECTIVE')
+        assert self.operator.SELECTIVE == SELECTIVE
+
+    def test_global(self):
+        assert hasattr(self.operator, 'GLOBAL')
+        assert self.operator.GLOBAL == GLOBAL
+
+    def test_include(self):
+        assert hasattr(self.operator, 'INCLUDE')
+        assert self.operator.INCLUDE == INCLUDE
+
+    def test_exclude(self):
+        assert hasattr(self.operator, 'EXCLUDE')
+        assert self.operator.EXCLUDE == EXCLUDE
